//
// Kernel Debugger Protocol Module - VMware Shared Memory
// Copyright (C) 2007 Ken Johnson (Skywing)
// http://www.nynaeve.net
// skywing@valhallalegends.com
//
// The kdvmware module is a simple driver that once loaded, performs an export
// table patch on the NTOS imports on kdcom.dll in order to gain control of the
// kernel debugger interface.  When sending or receiving a packet, we write
// bits to a shared memory region and then suspend the VM until the host-side
// debugger can process the event.  This is accomplished by repurposing the OS
// not loaded magic VMware backdoor API as a kernel debugger event notification
// mechanism.
//
// Host-side, the debugger engine has been patched to send and receive packets
// from the shared memory region.  VMware VMX is patched to gain control of the
// OS not loaded backdoor API.
//

#include <ntifs.h>
#include <Aux_klib.h>
#include <ntstrsafe.h>
#include "kdvmware.h"

#define VmkdPrintEx __noop

#define DISABLE_ON_BUGCHECK    0

typedef KDVM_DATA_BLOCK volatile * PKDVM_DATA_BLOCK_V;

PKDVM_DATA_BLOCK_V        KdVmDataBlock;
BOOLEAN                   KdVmInBugCheck;
extern PUCHAR            *KiBugCheckData;

VOID
KdVmEnterBugCheck(
	__in BOOLEAN InKeBugCheck2
	)
{
	KdVmInBugCheck = TRUE;

	KdVmNotifyEvent(
		KdVmOperationBugCheck
		);

	*KdDebuggerNotPresent              = TRUE;
	SharedUserData->KdDebuggerEnabled &= ~2;

	UNREFERENCED_PARAMETER( InKeBugCheck2 );
}

KD_RECV_CODE
KdReceivePacket(
	__in ULONG PacketType,
	__inout_opt PKD_BUFFER PacketData,
	__inout_opt PKD_BUFFER SecondaryData,
	__out_opt PULONG PayloadBytes,
	__inout_opt PKD_CONTEXT KdContext
	)
{
	if (!KdVmDataBlock || KdVmInBugCheck)
	{
		if (PacketType == PACKET_TYPE_KD_BREAKIN_POLL)
			return KD_RECV_CODE_TIMEOUT;
		else
			return KD_RECV_CODE_FAILED;
	}

#if DISABLE_ON_BUGCHECK
	//
	// Check to see if we are getting called during a bug check.  If so then
	// we will disable the debugger connection so that the bug check can be
	// processed.  This is to allow for debugging of the debugger link itself
	// if we get into a bad state as we'll be able to write a bug check dump.
	//

	if (*KiBugCheckData)
	{
		KdVmEnterBugCheck(TRUE);

		if (PacketType == PACKET_TYPE_KD_BREAKIN_POLL)
			return KD_RECV_CODE_TIMEOUT;
		else
			return KD_RECV_CODE_FAILED;
	}
#endif

	if (KdVmDataBlock->Parameters.PrimaryLength +
		KdVmDataBlock->Parameters.SecondaryLength > KD_MAX_DATA_BLOCK_SIZE)
	{
		KdVmEnterBugCheck(TRUE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_LARGE_RECV,
			(ULONG_PTR)PacketData,
			(ULONG_PTR)SecondaryData,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	//
	// Copy the request to the data block and fire it off.
	//

	KdVmDataBlock->Parameters.PacketType      = PacketType;
	KdVmDataBlock->Parameters.PrimaryLength   = PacketData ? PacketData->MaximumLength : 0;
	KdVmDataBlock->Parameters.SecondaryLength = SecondaryData ? SecondaryData->MaximumLength : 0;

	KdVmNotifyEvent(
		KdVmOperationReceivePacket
		);

	//
	// Copy the reply data to the caller.
	//

	if (KdVmDataBlock->Parameters.RecvCode == KD_RECV_CODE_OK)
	{
		if (PacketData)
		{
			PacketData->Length = KdVmDataBlock->Parameters.PrimaryLength;

			RtlCopyMemory(
				PacketData->Data,
				(PVOID)KdVmDataBlock->Parameters.Buffer,
				KdVmDataBlock->Parameters.PrimaryLength
				);
		}

		if (SecondaryData)
		{
			SecondaryData->Length = KdVmDataBlock->Parameters.SecondaryLength;

			RtlCopyMemory(
				SecondaryData->Data,
				(PVOID)(KdVmDataBlock->Parameters.Buffer + KdVmDataBlock->Parameters.PrimaryLength),
				KdVmDataBlock->Parameters.SecondaryLength
				);
		}

		if (PayloadBytes)
			*PayloadBytes = KdVmDataBlock->Parameters.SecondaryLength;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_BREAK_IN_REQUESTED)
	{
		if (KdContext)
			KdContext->BreakInRequested = TRUE;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_DISCONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = TRUE;
		SharedUserData->KdDebuggerEnabled &= ~2;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_CONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = FALSE;
		SharedUserData->KdDebuggerEnabled |= 2;
	}

	return KdVmDataBlock->Parameters.RecvCode;
}

VOID
KdSendPacket(
	__in ULONG PacketType,
	__in PKD_BUFFER PacketData,
	__in_opt PKD_BUFFER SecondaryData,
	__inout PKD_CONTEXT KdContext
	)
{
	if (!KdVmDataBlock || KdVmInBugCheck)
		return;

#if DISABLE_ON_BUGCHECK
	//
	// Check to see if we are getting called during a bug check.  If so then
	// we will disable the debugger connection so that the bug check can be
	// processed.  This is to allow for debugging of the debugger link itself
	// if we get into a bad state as we'll be able to write a bug check dump.
	//

	if (*KiBugCheckData)
	{
		KdVmEnterBugCheck(TRUE);

		return;
	}
#endif

	//
	// Copy the request to the data block and fire it off.
	//


	KdVmDataBlock->Parameters.PacketType      = PacketType;
	KdVmDataBlock->Parameters.PrimaryLength   = PacketData ? PacketData->Length : 0;
	KdVmDataBlock->Parameters.SecondaryLength = SecondaryData ? SecondaryData->Length : 0;

	if (KdVmDataBlock->Parameters.PrimaryLength +
		KdVmDataBlock->Parameters.SecondaryLength > KD_MAX_DATA_BLOCK_SIZE)
	{
		KdVmEnterBugCheck(FALSE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_LARGE_SEND,
			(ULONG_PTR)PacketData,
			(ULONG_PTR)SecondaryData,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	if (PacketData)
	{
		RtlCopyMemory(
			(PVOID)KdVmDataBlock->Parameters.Buffer,
			PacketData->Data,
			KdVmDataBlock->Parameters.PrimaryLength
			);
	}

	if (SecondaryData)
	{
		RtlCopyMemory(
			(PVOID)(KdVmDataBlock->Parameters.Buffer + KdVmDataBlock->Parameters.PrimaryLength),
			SecondaryData->Data,
			KdVmDataBlock->Parameters.SecondaryLength
			);
	}

	KdVmNotifyEvent(
		KdVmOperationSendPacket
		);

	//
	// Copy the reply data to the caller.
	//

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_BREAK_IN_REQUESTED)
	{
		if (KdContext)
			KdContext->BreakInRequested = TRUE;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_DISCONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = TRUE;
		SharedUserData->KdDebuggerEnabled &= ~2;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_CONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = FALSE;
		SharedUserData->KdDebuggerEnabled |= 2;
	}
}

NTSTATUS
KdSave(
	PVOID Unused
	)
{
	UNREFERENCED_PARAMETER( Unused );

	return STATUS_SUCCESS;
}

NTSTATUS
KdRestore(
	PVOID Unused
	)
{
	UNREFERENCED_PARAMETER( Unused );

	return STATUS_SUCCESS;
}

NTSTATUS
KdD0Transition(
	VOID
	)
{
	return STATUS_SUCCESS;
}

NTSTATUS
KdD3Transition(
	VOID
	)
{
	return STATUS_SUCCESS;
}

NTSTATUS
KdDebuggerInitialize0(
	__in PLOADER_PARAMETER_BLOCK LoaderParameterBlock
	)
{
	return STATUS_SUCCESS;
}

NTSTATUS
KdDebuggerInitialize1(
	__in PLOADER_PARAMETER_BLOCK LoaderParameterBlock
	)
{
	//
	// If the debugger is not enabled then we'll not try and connect.
	//

	if (!KD_DEBUGGER_ENABLED)
		return STATUS_SUCCESS;

	//
	// Connect to the host.
	//

	(VOID) KdVmInitializeHostCommunication();

	return STATUS_SUCCESS;
}

NTSTATUS
KdVmInitializeHostCommunication(
	VOID
	)
{
	//
	// Establish a communications link with the host.  This involves allocating
	// and initializing our memory region and trapping to the host until we get
	// a successful response.
	//

	NTSTATUS         Status;
	PHYSICAL_ADDRESS MaxAddr;
	ULONG            Tries;

	MaxAddr.LowPart  = 0xFFFFFFFF;
	MaxAddr.HighPart = 0xFFFFFFFF;

	Tries            = 0;

	KdVmDataBlock = (PKDVM_DATA_BLOCK_V)MmAllocateContiguousMemory(
		sizeof(KDVM_DATA_BLOCK),
		MaxAddr
		);

	if (!KdVmDataBlock)
	{
		VmkdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Failed to allocate contiguous physical memory.\n"));

		Status = STATUS_NO_MEMORY;
		return Status;
	}

	//
	// We need to initialize the KDVM data block.  Do so now.
	//

	RtlZeroMemory(
		(PVOID)KdVmDataBlock,
		sizeof(KDVM_DATA_BLOCK)
		);

	RtlCopyMemory(
		(PVOID)KdVmDataBlock->Magic,
		KDVM_SIGNATURE_INITIALIZING,
		KDVM_SIGNATURE_LENGTH
		);

	KdVmDataBlock->Magic[ 62 ] = 0xAA;

	KdVmDataBlock->Length         = sizeof(KDVM_DATA_BLOCK);

	//
	// We'll stop here until the host acknowledges us.
	//

	while (!KdVmDataBlock->HostLinkActive && Tries < 5)
	{
		VmkdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Trying to connect to host (%lu)...\n", Tries));

		Tries += 1;

		KdVmNotifyEvent(KdVmOperationInitialize);
	}

	if (!KdVmDataBlock->HostLinkActive)
	{
		VmkdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Failed to connect to host!  Unloading driver image.\n"));

		RtlZeroMemory(
			(PVOID)KdVmDataBlock,
			sizeof(KDVM_DATA_BLOCK)
			);

		MmFreeContiguousMemory((PVOID)KdVmDataBlock);
		KdVmDataBlock = 0;

		return STATUS_PORT_DISCONNECTED;
	}

	//
	// We're all good to go.  Let's return success.
	//

	return STATUS_SUCCESS;
}

VOID
KdVmNotifyEvent(
	KDVM_OPERATION Operation
	)
{
	//
	// Tell the host what operation we're performing for debugging and control
	// purposes.  This allows the host to know that we're still alive as we
	// unset the operation code after every transaction.
	//
	// We also unset the HostLinkActive flag for the duration of the
	// transaction so that we can know that the host side code processed our
	// request properly.  If they did not, then that's bad and we'll want to
	// reconnect the link.
	//

	KdVmDataBlock->Operation      = Operation;
	KdVmDataBlock->HostLinkActive = FALSE;

	KdVmpNotifyEvent();

//	while (KdVmDataBlock->Operation != KdVmOperationNone)
//		KeStallExecutionProcessor(100);

	//
	// If the host isn't linked anymore and we weren't already trying to do an
	// initialization operation, then try to reconnect once right now.
	//
	// N.B. There cannot be any debug prints here as any debug prints would
	// not succeed as the kernel debugger connection is down at this point.
	//

	if (!KdVmDataBlock->HostLinkActive && 
		Operation != KdVmOperationInitialize)
	{
		RtlCopyMemory(
			(PVOID)KdVmDataBlock->Magic,
			KDVM_SIGNATURE_INITIALIZING,
			KDVM_SIGNATURE_LENGTH
			);

		KdVmDataBlock->Magic[ 62 ] = 0xAA;

//		RtlStringCbCopyA(
//			(PCHAR)KdVmDataBlock->DebugString,
//			128,
//			"VM Host Disconnect!"
//			);

		KdVmNotifyEvent(
			KdVmOperationInitialize
			);
	}

	//
	// If we're requested to initiate a bug check, and the system is not
	// already processing a bug check, then initiate one now.
	//

	if (KdVmDataBlock->BugCheckRequested &&
		!KdVmInBugCheck                  &&
		Operation != KdVmOperationBugCheck)
	{
		KdVmEnterBugCheck(FALSE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_USER_REQUEST,
			(ULONG_PTR)*KiBugCheckData,
			(ULONG_PTR)0,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	KdVmDataBlock->Operation = KdVmOperationNone;
}
