//
// Virtual Machine Protocol Module - VMware Shared Memory
// Copyright (C) 2007 Ken Johnson (Skywing)
// http://www.nynaeve.net
// skywing@valhallalegends.com
//
// The vmxpatch.dll module implements the host side interface for the
// kdvmware.sys driver.  It creates a named pipe named off of the vmware-vmx
// current directory that can be connected to by DbgEng's KDCOM-over-PIPE
// support.
//
// For example:
//
// kd -k com:pipe,port=\\.\pipe\kdvmware_windowsserver2008
//
// (Assuming vmware-vmx.exe was running in directory "WindowsServer2008".)
//
// Use vmxinject.exe <vmware-vmx-pid> to inject the hook module into the
// target vmware-vmx.exe process.
//

#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <strsafe.h>
#include <process.h>
#include <stdarg.h>
#include "vmxpatch.h"

#ifndef _PSAPI_H

//
// WDK 6000 is lame and doesn't have PSAPI.h for some ridiculously stupid
// reason.
//

typedef struct _MODULEINFO {
    LPVOID lpBaseOfDll;
    DWORD SizeOfImage;
    LPVOID EntryPoint;
} MODULEINFO, *LPMODULEINFO;


BOOL
WINAPI
GetModuleInformation(
    __in HANDLE hProcess,
    __in HMODULE hModule,
    __out LPMODULEINFO lpmodinfo,
    __in DWORD cb
    );

#endif

unsigned
__stdcall
VmxPatchThreadRoutine(
	void *Arguments
	);

typedef
BOOLEAN
(__cdecl * IsVMInPrivilegedModeProc)(
	VOID
	);

HANDLE g_VmxPatchThread;
ULONG  g_VmxPatchThreadId;
HANDLE g_VmxDataFlushedEvent;
HANDLE g_VmxDataReadyEvent;
HANDLE g_VmxKdPipe = INVALID_HANDLE_VALUE;
HANDLE g_VmxKdPipeEvent;
CRITICAL_SECTION g_VmxDataLock;
UCHAR g_VmxDataBuf[ KD_MAX_DATA_BLOCK_SIZE ];
ULONG g_VmxDataLen;
volatile BOOLEAN g_LogPackets = FALSE;
volatile BOOLEAN g_ShutDown   = FALSE;
BOOLEAN g_PipeConnected       = FALSE;

PKDVM_DATA_BLOCK g_KdVmDataBlock;
KDVM_OPERATION g_LastKdVmOp = KdVmOperationNone;

IsVMInPrivilegedModeProc IsVMInPrivilegedMode;

USHORT OSNotFoundHandlerOpcodes[] =
{
	0x55, 0x8B, 0xEC, 0x81, 0xEC, 0x20, 0x03, 0x00, 0x00, 0x53, 0x32, 0xDB,
	0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x84, 0xC0
};

USHORT InPrivilegedModeOpcodes[] =
{
	0xA1, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x64, 0x8B, 0x0D,  0x2C, 0x00, 0x00, 0x00, 0x8B, 0x14, 0x81, 0x8B,
	0x82, 0xFC, 0x03, 0x00, 0x00, 0x8B, 0x15, 0xFFFF,  0xFFFF, 0xFFFF, 0xFFFF, 0xC1, 0xE0, 0x0C, 0x56, 0x8B, 
	0xB4, 0x10, 0x54, 0xC2, 0xFF, 0xFF, 0x33, 0xC9,  0x85, 0xF6, 0x0F, 0x94, 0xC1, 0x8A, 0xC1, 0x5E, 
	0xC3
};

int
__cdecl
VmxPatchOSNotFound(
	void
	);

void
VmxPatchDebugPrint(
	__in __format_string const char *Fmt,
	...
	);

ULONG
VmxPatchExceptionFilter(
	__in ULONG ExceptionCode
	);

void dumphex(const char *buf, int len, int pos);

BOOL
WINAPI
DllMain(
	__in HMODULE Module,
	__in ULONG Reason,
	__in_opt PCONTEXT Context OPTIONAL
	)
{
	WCHAR ModuleName[ MAX_PATH + 1 ];

	switch (Reason)
	{

	case DLL_PROCESS_ATTACH:
		//
		// We don't care about TLS callouts.
		//

		if (!GetModuleFileName(
			0,
			ModuleName,
			MAX_PATH))
			return FALSE;

		_wcsupr(ModuleName);

		DisableThreadLibraryCalls(Module);

		InitializeCriticalSection(&g_VmxDataLock);

		//
		// We only initialize if we're in VMWARE-VMX.
		//

		if (!wcsstr(ModuleName, L"VMWARE-VMX.EXE"))
		{
			VmxPatchDebugPrint(
				"DllMain(): Loaded in wrong process '%ws' (not enabling hooks)...\n",
				ModuleName);

			break;
		}

		VmxPatchDebugPrint(
			"DllMain(): Initializing vmware-vmx hooks.\n");

		g_VmxDataFlushedEvent = CreateEvent(
			0,
			FALSE,
			FALSE,
			0
			);

		if (!g_VmxDataFlushedEvent)
			return FALSE;

		g_VmxDataReadyEvent = CreateEvent(
			0,
			FALSE,
			FALSE,
			0
			);

		if (!g_VmxDataReadyEvent)
		{
			CloseHandle(g_VmxDataFlushedEvent);

			return FALSE;
		}

		//
		// We'll be wanting to create our thread to do the real work.
		//

		g_VmxPatchThread = (HANDLE)_beginthreadex(
			0,
			0,
			VmxPatchThreadRoutine,
			Module,
			0,
			&g_VmxPatchThreadId
			);

		if (g_VmxPatchThread == 0)
			return FALSE;
		break;

	case DLL_PROCESS_DETACH:
		if (g_VmxPatchThread)
			CloseHandle(g_VmxPatchThread);
		if (g_VmxDataReadyEvent)
			CloseHandle(g_VmxDataReadyEvent);
		if (g_VmxDataFlushedEvent)
			CloseHandle(g_VmxDataFlushedEvent);

		DeleteCriticalSection(&g_VmxDataLock);
		break;

	default:
		break;

	}

	return TRUE;
}

BOOLEAN
VmxPatchMatchSearchPattern(
	__in CONST UCHAR* Search,
	__in CONST USHORT* Pattern,
	__in SIZE_T Length
	)
{
	SIZE_T i;

	for (i = 0;
	     i < Length;
	     i += 1)
	{
		if (Pattern[ i ] == 0xFFFF)
			continue;

		if (Pattern[ i ] == Search[ i ])
			continue;

		return FALSE;
	}

	return TRUE;
}

BOOLEAN
VmxPatchHookOSNotFound(
	VOID
	)
{
	MODULEINFO ModuleInfo;
	PUCHAR     p;
	ULONG      i;
	HMODULE    Module;
	ULONG      OldProt;

	Module = GetModuleHandle(0);

	//
	// Get module bounds for vmware-vmx
	//

	if (!GetModuleInformation(
		GetCurrentProcess(),
		Module,
		&ModuleInfo,
		sizeof(MODULEINFO)))
	{
		VmxPatchDebugPrint(
			"VmxPatchHookOSNotFound(): GetModuleInformation fails - %lu\n",
			GetLastError());
		return FALSE;
	}

	//
	// Find our target, the IsVMInPrivilegedMode routine.
	//

	for (i = 0;
	     i < ModuleInfo.SizeOfImage;
	     i += 1)
	{
		p = ((UCHAR *)Module) + i;

		if (VmxPatchMatchSearchPattern(
			p,
			InPrivilegedModeOpcodes,
			sizeof(InPrivilegedModeOpcodes) / 2))
		{
			//
			// Save away our pointer to IsVMInPrivilegedMode.
			//

			IsVMInPrivilegedMode = (IsVMInPrivilegedModeProc)p;

			break;
		}
	}

	if (!IsVMInPrivilegedMode)
	{
		VmxPatchDebugPrint(
			"VmxPatchHookOSNotFound(): Failed to locate IsVMInPrivilegedMode (security check).\n"
			);

		return FALSE;
	}


	//
	// Find our target, the OS Not Found command handler.
	//

	for (i = 0;
	     i < ModuleInfo.SizeOfImage;
	     i += 1)
	{
		p = ((UCHAR *)Module) + i;

		if (VmxPatchMatchSearchPattern(
			p,
			OSNotFoundHandlerOpcodes,
			sizeof(OSNotFoundHandlerOpcodes) / 2))
		{
			//
			// Great, we've located the target.  Let's patch it to jump to our
			// code instead.
			//

			VirtualProtect(
				p,
				5,
				PAGE_EXECUTE_READWRITE,
				&OldProt
				);

			p[ 0 ] = 0xE9;
			*(PULONG_PTR)(p + 1) = (ULONG_PTR)VmxPatchOSNotFound - 
				(ULONG_PTR)(p + 5);

			VirtualProtect(
				p,
				5,
				OldProt,
				&OldProt
				);

			return TRUE;
		}
	}

	VmxPatchDebugPrint(
		"VmxPatchHookOSNotFound(): Failed to locate OSNotFound handler.\n"
		);

	return FALSE;
}

HANDLE
VmxPatchCreateNamedPipe(
	VOID
	)
{
	WCHAR   DirectoryName[ MAX_PATH + 1];
	WCHAR   PipeName[ MAX_PATH + 1];
	WCHAR  *VMName;

	//
	// Name the named pipe.  This would be better if we could get the VM name
	// but for now this works.
	//

	if (!GetCurrentDirectory(
		MAX_PATH,
		DirectoryName
		))
		StringCbCopy(
			DirectoryName,
			sizeof(DirectoryName),
			L"kdvmware");

	VMName = wcsrchr(
		DirectoryName,
		L'\\'
		);

	if (!VMName)
		VMName  = DirectoryName;
	else
		VMName += 1;
//PIPE_NAME
	if (FAILED(StringCbPrintf(
		PipeName,
		sizeof(PipeName),
		L"\\\\.\\pipe\\kdvmware_%s",
		VMName)))
		return 0;

	VmxPatchDebugPrint(
		"VmxPatchCreateNamedPipe(): Using pipe name '%ws'.\n",
		PipeName
		);

	return CreateNamedPipe(
		PipeName,
		PIPE_ACCESS_DUPLEX | FILE_FLAG_FIRST_PIPE_INSTANCE | FILE_FLAG_OVERLAPPED,
		PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT,
		1,
		1024,
		1024,
		10000,
		0
		);
}

unsigned
__stdcall
VmxPatchThreadRoutine(
	void *Arguments
	)
{
	HMODULE    Module;
	HANDLE     Pipe;
	UCHAR      Buf[ 8192 ];
//	UCHAR      Overflow[ 1024 ];
	ULONG      DataAvailable;
	ULONG      BytesRead;
	BOOLEAN    PipeConnected;
	BOOLEAN    BufferFull;
	ULONG      Limit;
	OVERLAPPED Overlapped;

	Module        = (HMODULE)Arguments;
	Pipe          = INVALID_HANDLE_VALUE;
	DataAvailable = 0;
	PipeConnected = FALSE;

	Overlapped.hEvent = 0;

	//
	// First, we'll need to find our patch location.
	//

	for (;;)
	{
		if (!(Pipe = VmxPatchCreateNamedPipe()))
		{
			VmxPatchDebugPrint(
				"VmxPatchThreadRoutine(): VmxPatchCreateNamedPipe fails, %lu.\n",
				GetLastError());
			break;
		}

		Overlapped.hEvent = CreateEvent(0, TRUE, FALSE, 0);

		if (!Overlapped.hEvent)
			break;

		g_VmxKdPipeEvent = CreateEvent(0, TRUE, FALSE, 0);

		if (!VmxPatchHookOSNotFound())
		{
			VmxPatchDebugPrint(
				"VmxPatchThreadRoutine(): VmxPatchHookOSNotFound failed.\n");
			break;
		}

		g_VmxKdPipe = Pipe;

		for (;;)
		{
			Sleep(1000);
		}
#if 0
		for (;;)
		{
			if (!PipeConnected)
			{
				Overlapped.Offset       = 0;
				Overlapped.OffsetHigh   = 0;
				Overlapped.Internal     = 0;
				Overlapped.InternalHigh = 0;

				if (!ConnectNamedPipe(
					Pipe,
					&Overlapped))
				{
					BOOL Success;

					if (GetLastError() == ERROR_IO_PENDING)
					{
						Success = GetOverlappedResult(
							Pipe,
							&Overlapped,
							&BytesRead,
							TRUE
							);
					}

					if (!Success && GetLastError() != ERROR_PIPE_CONNECTED)
					{
						Sleep(1000);
						continue;
					}
					else
					{
						PipeConnected = TRUE;

						VmxPatchDebugPrint(
							"VmxPatchThreadRoutine(): ConnectNamedPipe returns FALSE (ERROR_PIPE_CONNECTED), pipe is connected.\n");
					}
				}
				else
				{
					PipeConnected = TRUE;

					VmxPatchDebugPrint(
						"VmxPatchThreadRoutine(): ConnectNamedPipe returns TRUE, pipe is connected.\n");
				}

				ResetEvent(
					Overlapped.hEvent
					);
			}

			//
			// Receive bits from the kernel debugger connection.
			//

			Overlapped.Offset       = 0;
			Overlapped.OffsetHigh   = 0;
			Overlapped.Internal     = 0;
			Overlapped.InternalHigh = 0;

//			if (DataAvailable < sizeof( Buf ))
//			{
				if (!ReadFile(
					Pipe,
					Buf + DataAvailable,
					sizeof( Buf ) - DataAvailable,
					&BytesRead,
					&Overlapped))
				{
					if ((GetLastError() != ERROR_IO_PENDING) ||
						(!GetOverlappedResult(
							Pipe,
							&Overlapped,
							&BytesRead,
							TRUE)))
					{
						VmxPatchDebugPrint(
							"VmxPatchThreadRoutine(): ReadFile fails, %lu\n",
							GetLastError());
						DisconnectNamedPipe(Pipe);
						PipeConnected = FALSE;
						DataAvailable = 0;
						continue;
					}
				}

				ResetEvent(
					Overlapped.hEvent
					);

				DataAvailable += BytesRead;

				//
				// Grab the lock synchronizing access to the staging buffer.
				//

				EnterCriticalSection(&g_VmxDataLock);

				//
				// Copy as much data as possible to the staging buffer.
				//

				Limit = sizeof( g_VmxDataBuf ) - g_VmxDataLen;
				Limit = min(Limit, DataAvailable);

				memcpy(
					g_VmxDataBuf + g_VmxDataLen,
					Buf,
					Limit
					);

	//			VmxPatchDebugPrint(
	//				"VmxPatchThreadRoutine(): Copied %lu bytes to staging buffer.\n",
	//				Limit
	//				);

				g_VmxDataLen += Limit;

				LeaveCriticalSection(&g_VmxDataLock);

				//
				// Remove any data that we've copied to the staging buffer from our
				// read buffer.
				//

				DataAvailable -= Limit;

				memmove(
					Buf,
					Buf + Limit,
					DataAvailable
					);

				if (Limit)
					SetEvent(g_VmxDataReadyEvent);
//			}
/*			else
			{
				if (!ReadFile(
					Pipe,
					Overflow,
					sizeof( Overflow ),
					&BytesRead,
					&Overlapped)                         &&
					(GetLastError() == ERROR_IO_PENDING) &&
					(!GetOverlappedResult(
						Pipe,
						&Overlapped,
						&BytesRead,
						TRUE)))
				{
					VmxPatchDebugPrint(
						"VmxPatchThreadRoutine(): ReadFile fails, %lu\n",
						GetLastError());
					DisconnectNamedPipe(Pipe);
					PipeConnected = FALSE;
					DataAvailable = 0;
					continue;
				}

				ResetEvent(
					Overlapped.hEvent
					);

				VmxPatchDebugPrint(
					"VmxPatchThreadRoutine(): Drop %lu bytes from kernel debugger (UART overflow)\n",
					BytesRead);
			}
*/
			//
			// If we're out of buffer space, then wait for data to be removed
			// from the staging buffer before we try another read.
			//
#if 1
			while (DataAvailable == sizeof( Buf ))
			{
				VmxPatchDebugPrint("VmxPatchThreadRoutine(): Out of buffer space - waiting for VM to read data\n");

				//
				// Wait for the VM to pull some data from our staging buffer.
				//

				WaitForSingleObject(g_VmxDataFlushedEvent, INFINITE);

				//
				// Acquire the lock for the staging buffer data.
				//

				EnterCriticalSection(&g_VmxDataLock);

				//
				// Copy as much data as possible to the staging buffer.
				//

				Limit = sizeof( g_VmxDataBuf ) - g_VmxDataLen;
				Limit = min(Limit, DataAvailable);

				memcpy(
					g_VmxDataBuf + g_VmxDataLen,
					Buf,
					Limit
					);

				g_VmxDataLen += Limit;

				LeaveCriticalSection(&g_VmxDataLock);

				//
				// Remove any data that we've copied to the staging buffer from our
				// read buffer.
				//

				DataAvailable -= Limit;

				memmove(
					Buf,
					Buf + Limit,
					DataAvailable
					);

				if (Limit)
					SetEvent(g_VmxDataReadyEvent);
			}
#endif
		}
#endif
	}

	VmxPatchDebugPrint(
		"VmxPatchThreadRoutine(): Thread is exiting.\n"
		);

	if (Pipe != INVALID_HANDLE_VALUE)
	{
		CloseHandle(Pipe);

		g_VmxKdPipe = INVALID_HANDLE_VALUE;
	}

	if (g_VmxKdPipeEvent)
		CloseHandle(g_VmxKdPipeEvent);

	if (Overlapped.hEvent)
		CloseHandle(Overlapped.hEvent);

	FreeLibraryAndExitThread(
		Module,
		0
		);
}

BOOLEAN
VmxPatchFindKdVmDataBlock(
	VOID
	)
{
	MEMORY_BASIC_INFORMATION mbi;
	PUCHAR                   p, q;

	p = (PUCHAR)0x0;

	//
	// Scan the address space for our signature.  This is lame and would be
	// replaced with some guaranteed way to communicate securely with the guest
	// in an implementation that had the cooperation of the VMM software...
	//
	// However, since I don't have any way to get help from the VMware people
	// about their internal APIs, this is the best I can do.  It works in a
	// pinch.
	//

	while (VirtualQuery(
		p,
		&mbi,
		sizeof( MEMORY_BASIC_INFORMATION )))
	{
		q  = p;
		p += mbi.RegionSize;

		//
		// VM physical pages must be committed.
		//

		if (!(mbi.State & MEM_COMMIT))
			continue;

		//
		// We don't want NOACCESS or GUARD pages.
		//

		if (mbi.Protect & (PAGE_NOACCESS | PAGE_GUARD))
			continue;

		//
		// We don't want pages that back a .exe or .dll.
		//

		if (mbi.Type & MEM_IMAGE)
			continue;

		//
		// We only want pages that are writable.
		//

		if (!(mbi.Protect & (PAGE_READWRITE | PAGE_EXECUTE_READWRITE | PAGE_WRITECOPY | PAGE_EXECUTE_WRITECOPY)))
			continue;

		//
		// Okay, this region might have something of value on it.  Let's go and
		// search it.
		//
		// N.B. If I knew for sure that VMware allocated guest physical pages
		// on a host physical page boundary, then we could speed the search
		// considerably by only checking page boundaries, as the driver only
		// uses newly allocated page sized or larger allocations that are
		// physically contiguous.
		//
		// N.B. We currently assume that if VMware guest physical memory pages
		// are contiguous, then the virtual pages backing them in the
		// vmware-vmx process are going to be virtually contiguous.
		//


		while ((q + KDVM_SIGNATURE_LENGTH) < p)
		{
			//
			// Check to see that we match the special initialization
			// signature.
			//

			if (!memcmp(q, KDVM_SIGNATURE_INITIALIZING, KDVM_SIGNATURE_LENGTH))
			{
				//
				// Okay, we've got a probable match.  Let's return it if the
				// size also checks out.
				//

				g_KdVmDataBlock = (PKDVM_DATA_BLOCK)q;

				if (g_KdVmDataBlock->Length != sizeof(KDVM_DATA_BLOCK))
				{
					VmxPatchDebugPrint(
						"VmxPatchFindKdVmDataBlock(): Found virtual address in vmware-vmx for probable KDVM_DATA_BLOCK (%p), but length was wrong (%lu).\n",
						q,
						g_KdVmDataBlock->Length
						);

					g_KdVmDataBlock = 0;

					q += sizeof(KDVM_DATA_BLOCK);
					continue;
				}

				if (g_KdVmDataBlock->Operation != KdVmOperationInitialize)
				{
					VmxPatchDebugPrint(
						"VmxPatchFindKdVmDataBlock(): Found virtual address in vmware-vmx for probable KDVM_DATA_BLOCK (%p), but operation was wrong (%x).\n",
						q,
						(unsigned long)g_KdVmDataBlock->Operation
						);

					g_KdVmDataBlock = 0;

					q += sizeof(KDVM_DATA_BLOCK);
					continue;
				}

				VmxPatchDebugPrint(
					"VmxPatchFindKdVmDataBlock(): Found virtual address in vmware-vmx for KDVM_DATA_BLOCK: %p\n",
					q
					);

				//
				// Let's save away the current operation.  If we get another
				// call and the last and current operations were both
				// initialize, then we've got the wrong block and there are
				// possibly multiple copies.  We've already eliminated the
				// current block by virtue of changing the signature, so we
				// can hope that the next time around, we'll get the right
				// block if this turns out to happen.
				//

				g_LastKdVmOp = g_KdVmDataBlock->Operation;

				//
				// Change the signature to that of a connected block.
				//

				memcpy(
					g_KdVmDataBlock->Magic,
					KDVM_SIGNATURE_ESTABLISHED,
					KDVM_SIGNATURE_LENGTH
					);

				//
				// Tell the driver that we've linked up with it host-side.
				//

				g_KdVmDataBlock->HostLinkActive = TRUE;

				return TRUE;
			}

			q += 1;
		}
	}

	return FALSE;
}

int
__cdecl
VmxPatchOSNotFound(
	void
	)
{
	//
	// Let's make sure the caller is really kernel mode first off.  However, a
	// malicious user could still hope to place their signatured block before
	// us and cause us to potentially overwrite another page with badness.  A
	// better implementation could be done with actual help from the VMM and
	// not as a patched-on hack.
	//
	// Because the attacker cannot control the contents overwritten, it would
	// be difficult to use this to do more than crash a system that is being
	// debugged with the fast KD API.  Thus I am considering it an acceptable
	// risk in the initial release.
	//

	if (!IsVMInPrivilegedMode())
		return -1;

	//
	// N.B. Race condition if VM got reset here, although I'm not even sure if
	// that is permitted to happen during a backdoor IO port call.
	//

	__try
	{
		//
		// First thing, verify that we've got a good connection to the guest.
		//

		if (!g_KdVmDataBlock                                                           ||
			memcmp(g_KdVmDataBlock, KDVM_SIGNATURE_ESTABLISHED, KDVM_SIGNATURE_LENGTH) ||
			g_KdVmDataBlock->Operation == KdVmOperationNone                            ||
			(g_LastKdVmOp == KdVmOperationInitialize                                   && 
			 g_KdVmDataBlock->Operation == KdVmOperationInitialize))
		{
			if (g_KdVmDataBlock)
			{
				VmxPatchDebugPrint(
					"VmxPatchOSNotFound(): KDVM data block appears to be invalid or the driver has violated the communication protocol, breaking the connection.\n");
			}

			g_KdVmDataBlock = 0;

			//
			// We weren't connected to the guest or the shared memory block has
			// been moved around.  We'll have to go searching for it.
			//

			if (!VmxPatchFindKdVmDataBlock())
				return -1;

			//
			// Because we just found the data block, it was initializing and we
			// should just return right away.
			//

			KdpResetState();

			return -1;
		}
		else
		{
			KD_BUFFER  PacketData;
			KD_BUFFER  SecondaryData;
			ULONG      PayloadBytes;
			KD_CONTEXT Context;



			g_LastKdVmOp = g_KdVmDataBlock->Operation;

			//
			// Tell the driver that we're still alive.
			//

			g_KdVmDataBlock->HostLinkActive = TRUE;

			//
			// We've got what appears to be a live connection to the guest-side
			// driver.  Let's transfer that data....
			//

			VmxPatchDebugPrint(
				"VmxPatchOSNotFound(): Servicing request %x.\n",
				(unsigned long)g_KdVmDataBlock->Operation
				);

			//
			// Set up our common parameters for SEND KD PACKET or RECEIVE KD
			// PACKET operations.
			//

			PacketData.Data      = g_KdVmDataBlock->Parameters.Buffer;
			PacketData.Length    = 0;
			SecondaryData.Data   = g_KdVmDataBlock->Parameters.Buffer + g_KdVmDataBlock->Parameters.PrimaryLength;
			SecondaryData.Length = 0;

			ZeroMemory(
				&Context.Overlapped,
				sizeof( OVERLAPPED )
				);

			g_KdVmDataBlock->Parameters.Flags = 0;

			Context.RetryCount       = 20;
			Context.BreakInRequested = FALSE;
			Context.Pipe             = g_VmxKdPipe;
			Context.Flags            = &g_KdVmDataBlock->Parameters.Flags;

			//
			// Convert the request into a call to the actual implementation
			// for the requested function, perform the operation, and return
			// the results to the requestor.
			//

			switch (g_KdVmDataBlock->Operation)
			{

			case KdVmOperationReceivePacket:
				{
					PacketData.MaximumLength    = g_KdVmDataBlock->Parameters.PrimaryLength;
					SecondaryData.MaximumLength = g_KdVmDataBlock->Parameters.SecondaryLength;

					g_KdVmDataBlock->Parameters.RecvCode = KdReceivePacket(
						g_KdVmDataBlock->Parameters.PacketType,
						&PacketData,
						&SecondaryData,
						&PayloadBytes,
						&Context
						);

					g_KdVmDataBlock->Parameters.PrimaryLength   = PacketData.Length;
					g_KdVmDataBlock->Parameters.SecondaryLength = SecondaryData.Length;
				}
				break;

			case KdVmOperationSendPacket:
				{
					PacketData.Length    = g_KdVmDataBlock->Parameters.PrimaryLength;
					SecondaryData.Length = g_KdVmDataBlock->Parameters.SecondaryLength;

					KdSendPacket(
						g_KdVmDataBlock->Parameters.PacketType,
						&PacketData,
						&SecondaryData,
						&Context
						);
				}
				break;

			case KdVmOperationBugCheck:
				{
					VmxPatchDebugPrint(
						"VmxPatchOSNotFound(): Guest has entered KeBugCheckEx, debugger link is disabled guest-side until reboot.  Shared memory ptr = %p\n",
						g_KdVmDataBlock
						);
				}
				break;

			}

			if (Context.BreakInRequested)
				g_KdVmDataBlock->Parameters.Flags |= KDVM_FLAG_BREAK_IN_REQUESTED;

			Sleep(0);

			return -1;
		}
	}
	__except(VmxPatchExceptionFilter(GetExceptionCode()))
	{
		return -1;
	}
}

void
VmxPatchDebugPrint(
	__in __format_string const char *Fmt,
	...
	)
{
	char    Buf[ 1024 ];
	va_list Ap;

	if (!IsDebuggerPresent())
		return;

	va_start(Ap, Fmt);

	StringCbVPrintfA(
		Buf,
		sizeof(Buf),
		Fmt,
		Ap
		);

	va_end(Ap);

	OutputDebugStringA(
		Buf
		);
}

ULONG
VmxPatchExceptionFilter(
	__in ULONG ExceptionCode
	)
{
	switch (ExceptionCode)
	{

	case EXCEPTION_ACCESS_VIOLATION:
	case EXCEPTION_IN_PAGE_ERROR:
		return EXCEPTION_EXECUTE_HANDLER;

	default:
		return EXCEPTION_CONTINUE_SEARCH;
		break;

	}
}

VOID
VmxPatch(
	VOID
	)
{
}


void dumphex(const char *buf, int len, int pos)
{
	int i, j;

	if (!g_LogPackets)
		return;

	for(j = 0; j < len; j += 16) {
		for(i = 0; i < 16; i++) {
			if(i + j < len)
				VmxPatchDebugPrint("%02x%c", (unsigned char)buf[i + j], j + i + 1 == pos ? '*' : ' ');
			else 
				VmxPatchDebugPrint("   ");
		}
		for(i = 0; i < 16; i++) {
			if(i + j < len)
				VmxPatchDebugPrint("%c", buf[i + j] >= ' ' ? buf[i + j] : '.');
			else 
				VmxPatchDebugPrint(" ");
		}
		VmxPatchDebugPrint("\n");
	}
}
